@@ -18,7 +18,8 @@
 # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
 
-"""Query Debian's Bug Tracking System (BTS).
+"""
+Query Debian's Bug Tracking System (BTS).
 
 This module provides a layer between Python and Debian's BTS. It
 provides methods to query the BTS using the BTS' SOAP interface, and the
@@ -28,7 +29,8 @@ Bugreport class which represents a bugreport from the BTS.
 
 from __future__ import division, unicode_literals, absolute_import, print_function
 
-import email
+import base64
+import email.feedparser
 from datetime import datetime
 import os
 import sys
@@ -42,7 +44,11 @@ if os.path.isdir(ca_path):
     os.environ['SSL_CERT_DIR'] = ca_path
 
 
-__version__ = '2.2'
+# please follow the semver semantics, i.e. MAJOR.MINOR.PATCH where
+# MAJOR: incompatible API changes
+# MINOR: add backwards-compatible functionality
+# PATCH: backwards-compatible bug fixes.
+__version__ = '2.6.1'
 
 
 PY2 = sys.version_info.major == 2
@@ -55,42 +61,66 @@ BTS_URL = 'https://bugs.debian.org/'
 # Max number of bugs to send in a single get_status request
 BATCH_SIZE = 500
 
-soap_client = SoapClient(location=URL, namespace=NS, soap_ns='soap')
 
 
 class Bugreport(object):
     """Represents a bugreport from Debian's Bug Tracking System.
 
     A bugreport object provides all attributes provided by the SOAP
-    interface. Most of the attributs are strings, the others are marked.
+    interface. Most of the attributes are strings, the others are
+    marked.
 
-    * bug_num: The bugnumber (int)
-    * severity: Severity of the bugreport
-    * tags: List of tags of the bugreport (list of strings)
-    * subject:  The subject/title of the bugreport
-    * originator: Submitter of the bugreport
-    * mergedwith: List of bugnumbers this bug was merged with (list of
-      ints)
-    * package: Package of the bugreport
-    * source: Source package of the bugreport
-    * date: Date of bug creation (datetime)
-    * log_modified: Date of update of the bugreport (datetime)
-    * done: Is the bug fixed or not (bool)
-    * archived: Is the bug archived or not (bool)
-    * unarchived: Was the bug unarchived or not (bool)
-    * fixed_versions: List of versions, can be empty even if bug is
-      fixed (list of strings)
-    * found_versions: List of version numbers where bug was found (list
-      of strings)
-    * forwarded: A URL or email address
-    * blocks: List of bugnumbers this bug blocks (list of ints)
-    * blockedby: List of bugnumbers which block this bug (list of ints)
-    * pending: Either 'pending' or 'done'
-    * msgid: Message ID of the bugreport
-    * owner: Who took responsibility for fixing this bug
-    * location: Either 'db-h' or 'archive'
-    * affects: List of Packagenames (list of strings)
-    * summary: Arbitrary text
+    Attributes
+    ----------
+
+    bug_num : int
+        The bugnumber
+    severity : str
+        Severity of the bugreport
+    tags : list of strings
+        Tags of the bugreport
+    subject : str
+        The subject/title of the bugreport
+    originator : str
+        Submitter of the bugreport
+    mergedwith : list of ints
+        List of bugnumbers this bug was merged with
+    package : str
+        Package of the bugreport
+    source : str
+        Source package of the bugreport
+    date : datetime
+        Date of bug creation
+    log_modified : datetime
+        Date of update of the bugreport
+    done : boolean
+        Is the bug fixed or not
+    archived : bool
+        Is the bug archived or not
+    unarchived : bool
+        Was the bug unarchived or not
+    fixed_versions : list of strings
+        List of versions, can be empty even if bug is fixed
+    found_versions : list of strings
+        List of version numbers where bug was found
+    forwarded : str
+        A URL or email address
+    blocks: list of ints
+        List of bugnumbers this bug blocks
+    blockedby : list of int
+        List of bugnumbers which block this bug
+    pending : str
+        Either 'pending' or 'done'
+    msgid : str
+        Message ID of the bugreport
+    owner : str
+        Who took responsibility for fixing this bug
+    location : str
+        Either 'db-h' or 'archive'
+    affects : list of str
+        List of Packagenames
+    summary : str
+        Arbitrary text
     """
 
     def __init__(self):
@@ -127,7 +157,7 @@ class Bugreport(object):
         # self.id = None
 
     def __unicode__(self):
-        s = '\n'.join('{}: {}'.format(key, str(value))
+        s = '\n'.join('{}: {}'.format(key, value)
                        for key, value in self.__dict__.items())
         return s + '\n'
 
@@ -140,7 +170,7 @@ class Bugreport(object):
     def __lt__(self, other):
         """Compare a bugreport with another.
 
-        The more open and and urgent a bug is, the greater the bug is:
+        The more open and urgent a bug is, the greater the bug is:
 
             outstanding > resolved > archived
 
@@ -191,7 +221,21 @@ class Bugreport(object):
 
 
 def get_status(*nrs):
-    """Returns a list of Bugreport objects."""
+    """Returns a list of Bugreport objects.
+
+    Given a list of bugnumbers this method returns a list of Bugreport
+    objects.
+
+    Parameters
+    ----------
+    nrs : int or list of ints
+        the bugnumbers
+
+    Returns
+    -------
+    bugs : list of Bugreport objects
+
+    """
     # If we called get_status with one single bug, we get a single bug,
     # if we called it with a list of bugs, we get a list,
     # No available bugreports returns an empty list
@@ -203,6 +247,7 @@ def get_status(*nrs):
         else:
             list_.append(nr)
     # Process the input in batches to avoid hitting resource limits on the BTS
+    soap_client = _build_soap_client()
     for i in range(0, len(list_), BATCH_SIZE):
         slice_ = list_[i:i + BATCH_SIZE]
         # I build body by hand, pysimplesoap doesn't generate soap Arrays
@@ -217,10 +262,20 @@ def get_status(*nrs):
 
 
 def get_usertag(email, *tags):
-    """Return a dictionary of "usertag" => buglist mappings.
+    """Get buglists by usertags.
+
+    Parameters
+    ----------
+    email : str
+    tags : tuple of strings
+        If tags are given the dictionary is limited to the matching
+        tags, if no tags are given all available tags are returned.
+
+    Returns
+    -------
+    mapping : dict
+        a mapping of usertag -> buglist
 
-    If tags are given the dictionary is limited to the matching tags, if
-    no tags are given all available tags are returned.
     """
     reply = _soap_client_call('get_usertag', email, *tags)
     map_el = reply('s-gensym3')
@@ -243,29 +298,39 @@ def get_usertag(email, *tags):
 
 
 def get_bug_log(nr):
-    """Return a list of Buglogs.
+    """Get Buglogs.
 
     A buglog is a dictionary with the following mappings:
-        "header" => string
-        "body" => string
-        "attachments" => list
-        "msg_num" => int
-        "message" => email.message.Message
+        * "header" => string
+        * "body" => string
+        * "attachments" => list
+        * "msg_num" => int
+        * "message" => email.message.Message
+
+    Parameters
+    ----------
+    nr : int
+        the bugnumber
+
+    Returns
+    -------
+    buglogs : list of dicts
+
     """
     reply = _soap_client_call('get_bug_log', nr)
     items_el = reply('soapenc:Array')
     buglogs = []
     for buglog_el in items_el.children():
         buglog = {}
-        buglog["header"] = _uc(str(buglog_el("header")))
-        buglog["body"] = _uc(str(buglog_el("body")))
+        buglog["header"] = _parse_string_el(buglog_el("header"))
+        buglog["body"] = _parse_string_el(buglog_el("body"))
         buglog["msg_num"] = int(buglog_el("msg_num"))
         # server always returns an empty attachments array ?
         buglog["attachments"] = []
 
         mail_parser = email.feedparser.FeedParser()
-        mail_parser.feed(str(buglog_el("header")))
-        mail_parser.feed(str(buglog_el("body")))
+        mail_parser.feed(buglog["header"])
+        mail_parser.feed(buglog["body"])
         buglog["message"] = mail_parser.close()
 
         buglogs.append(buglog)
@@ -273,38 +338,80 @@ def get_bug_log(nr):
 
 
 def newest_bugs(amount):
-    """Returns a list of bugnumbers of the `amount` newest bugs."""
+    """Returns the newest bugs.
+
+    This method can be used to query the BTS for the n newest bugs.
+
+    Parameters
+    ----------
+    amount : int
+        the number of desired bugs. E.g. if `amount` is 10 the method
+        will return the 10 latest bugs.
+
+    Returns
+    -------
+    bugs : list of int
+        the bugnumbers
+
+    """
     reply = _soap_client_call('newest_bugs', amount)
     items_el = reply('soapenc:Array')
     return [int(item_el) for item_el in items_el.children() or []]
 
 
 def get_bugs(*key_value):
-    """Returns a list of bugnumbers, that match the conditions given by
-    the key-value pair(s).
+    """Get list of bugs matching certain criteria.
+
+    The conditions are defined by key value pairs.
 
     Possible keys are:
-        "package": bugs for the given package
-        "submitter": bugs from the submitter
-        "maint": bugs belonging to a maintainer
-        "src": bugs belonging to a source package
-        "severity": bugs with a certain severity
-        "status": can be either "done", "forwarded", or "open"
-        "tag": see http://www.debian.org/Bugs/Developer#tags for
-            available tags
-        "owner": bugs which are assigned to `owner`
-        "bugs": takes list of bugnumbers, filters the list according to
-            given criteria
-        "correspondent": bugs where `correspondent` has sent a mail to
+        * "package": bugs for the given package
+        * "submitter": bugs from the submitter
+        * "maint": bugs belonging to a maintainer
+        * "src": bugs belonging to a source package
+        * "severity": bugs with a certain severity
+        * "status": can be either "done", "forwarded", or "open"
+        * "tag": see http://www.debian.org/Bugs/Developer#tags for
+           available tags
+        * "owner": bugs which are assigned to `owner`
+        * "bugs": takes single int or list of bugnumbers, filters the list
+           according to given criteria
+        * "correspondent": bugs where `correspondent` has sent a mail to
+
+    Arguments
+    ---------
+    key_value : str
+
+    Returns
+    -------
+    bugs : list of ints
+        the bugnumbers
+
+    Examples
+    --------
+    >>> get_bugs('package', 'gtk-qt-engine', 'severity', 'normal')
+    [12345, 23456]
 
-    Example: get_bugs('package', 'gtk-qt-engine', 'severity', 'normal')
     """
     # previous versions also accepted get_bugs(['package', 'gtk-qt-engine', 'severity', 'normal'])
     # if key_value is a list in a one elemented tuple, remove the
     # wrapping list
     if len(key_value) == 1 and isinstance(key_value[0], list):
         key_value = tuple(key_value[0])
-    reply = _soap_client_call('get_bugs', *key_value)
+
+    # pysimplesoap doesn't generate soap Arrays without using wsdl
+    # I build body by hand, converting list to array and using standard
+    # pysimplesoap marshalling for other types
+    method_el = SimpleXMLElement('<get_bugs></get_bugs>')
+    for arg_n, kv in enumerate(key_value):
+        arg_name = 'arg' + str(arg_n)
+        if isinstance(kv, (list, tuple)):
+            _build_int_array_el(arg_name, method_el, kv)
+        else:
+            method_el.marshall(arg_name, kv)
+
+    soap_client = _build_soap_client()
+    reply = soap_client.call('get_bugs', method_el)
     items_el = reply('soapenc:Array')
     return [int(item_el) for item_el in items_el.children() or []]
 
@@ -317,7 +424,7 @@ def _parse_status(bug_el):
     for field in ('originator', 'subject', 'msgid', 'package', 'severity',
                   'owner', 'summary', 'location', 'source', 'pending',
                   'forwarded'):
-        setattr(bug, field, _uc(str(bug_el(field))))
+        setattr(bug, field, _parse_string_el(bug_el(field)))
 
     bug.date = datetime.utcfromtimestamp(float(bug_el('date')))
     bug.log_modified = datetime.utcfromtimestamp(float(bug_el('log_modified')))
@@ -346,9 +453,38 @@ def _parse_status(bug_el):
     return bug
 
 
+# to support python 3.4.3, when using httplib2 as pysimplesoap transport we must
+# work around a bug in httplib2, which uses http.client.HTTPSConnection with
+# check_hostname=True, but with an empty ssl context that prevents the
+# certificate verification. Passing `cacert` to httplib2 through pysimplesoap
+# permits to create a valid ssl context.
+_soap_client_kwargs = {'location': URL, 'namespace': NS, 'soap_ns': 'soap'}
+if sys.version_info.major == 3 and sys.version_info < (3, 4, 3):
+    try:
+        from httplib2 import CA_CERTS
+    except ImportError:
+        pass
+    else:
+        _soap_client_kwargs['cacert'] = CA_CERTS
+
+def _build_soap_client():
+    """Factory method that creates a SoapClient.
+
+    For thread-safety we create SoapClients on demand instead of using a
+    module-level one.
+
+    Returns
+    -------
+    sc : SoapClient instance
+
+    """
+    return SoapClient(**_soap_client_kwargs)
+
+
 def _soap_client_call(method_name, *args):
-    # wrapper to work around pysimplesoap bug
-    # https://github.com/pysimplesoap/pysimplesoap/issues/31
+    """wrapper to work around a pysimplesoap bug
+    https://github.com/pysimplesoap/pysimplesoap/issues/31"""
+    soap_client = _build_soap_client()
     soap_args = []
     for arg_n, arg in enumerate(args):
         soap_args.append(('arg' + str(arg_n), arg))
@@ -375,7 +511,20 @@ def _parse_bool(el):
     return not value.strip() in ('', '0')
 
 
-"""Convert string to unicode.
+def _parse_string_el(el):
+    """read a string element, maybe encoded in base64"""
+    value = str(el)
+    el_type = el.attributes().get('xsi:type')
+    if el_type and el_type.value == 'xsd:base64Binary':
+        value = base64.b64decode(value)
+        if not PY2:
+            value = value.decode('utf-8', errors='replace')
+    value = _uc(value)
+    return value
+
+
+"""
+Convert string to unicode.
 
 This method only exists to unify the unicode conversion in this module.
 """
